/*******************************************
 *
 * Drehzahlsteuerung fr Schmiedeesse
 *
 * (C) Fingers elektrische Welt 2005
 *
 * Prozessor : PIC16F76
 *
 *******************************************/

#include	<pic.h>
#include	<stdlib.h>
#include  <stdio.h>
#include	<string.h>
#include  <math.h>

__CONFIG(0x3f32);

//---------------------------------------
//  Configuration word:
//	Bit 13,12	Code protection		11	=OFF
//	Bit 11,10	Code protection		11	=OFF
//	Bit 9,8		Code protection		11	=OFF
//	Bit 7		not implemented		x
//	Bit 6		brown out			0	=disabled
//	Bit 5,4		Code protection		11	=OFF
//	Bit	3		Power up timer		0	=enabled
//	Bit 2		Watchdog			0	= disabled
//	Bit 1,0		Oszillator			10	=HS
//----------------------------------------

#define PulsMax				25											// Maximale Anzahl von Perioden
#define FALSE          0
#define TRUE           1

volatile char	Schaltzustand = FALSE;					// Motor eingeschaltet oder nicht
volatile char SollPuls = PulsMax-1;						// Anzahl der eingeschalteten Perioden (Sollwert)
volatile char SollPause = 1;									// Anzahl der ausgeschalteten Perioden (Sollwert)
volatile char Puls = 0;                       // Anzahl der eingeschalteten Perioden (Istwert)
volatile char Pause = 0;                      // Anzahl der ausgeschalteten Perioden (Istwert)
unsigned int Drehzahl  = 0;                   // Via Capture gemessene Drehzahl

char outpbuf [30];                            // Universeller Puffer

char Sekundenticker = 0;

volatile unsigned int Potiwert;

//---------------------
// Watchdog rcksetzen
//---------------------
void WDR (void)
 {
  #asm
  clrwdt
  #endasm
 }

//------------------------------
// Einfache Verzgerungsroutine
//------------------------------
void Delay (unsigned int time)
 {
   unsigned char ttt;

   while (time > 0)
   	{
   	 ttt = 255;
   	 while (ttt>0) { ttt--; }
   	 WDR ();
   	 time-- ;
   	}

 }

//------------------------------------------------------
// Liefert einen zum Potiwert proportionalen Zahlenwert
//------------------------------------------------------
unsigned int GetPoti (void)
 {
   unsigned int zaehler = 0;                  // Zhlvariable

   TRISB &= ~0x80;                            // Portpin auf Ausgang schalten
   PORTB |= 0x80;                             // Portpin auf 1 Setzen um C zu entladen
   Delay (1000);                              // Kurze Pause
   TRISB |= 0x80;                             // Portpin auf Eingang schalten

   while (PORTB & 0x80)                       // Warten bis er auf Null geht, dabei Zeit messen
    {
     zaehler++;
     NOP();
     NOP();
     NOP();
     NOP();
     WDR ();
    }

   return (zaehler / 6);                     // Wert zurckliefern (ca. Prozent)
 }


//-------------------------
// 1 Byte via RS232 senden
//-------------------------
/*
void SendChar (unsigned char data)
{
 RP0=0;			            // Page 0 whlen
 while (TXIF != 1) {;}	// Sehen, ob Senderegister leer ist
 TXREG = data;			    //  Daten ins Senderegister
 WDR ();
}
*/

//---------------
// String senden
//---------------
void SendString (const char *data)
 {
 	do
 	 {
    while (TXIF != 1) {;}	// Sehen, ob Senderegister leer ist
   	TXREG = *data;        // Zeichen senden
 	 } while ( *data++ );

 }

//------------------------
// Schaltet den Motor ein
//------------------------
void SpannungEin (void)
 {
  PORTB &= ~0x40;
 }

//------------------------
// Schaltet den Motor aus
//------------------------
void SpannungAus (void)
 {
  PORTB |= 0x40;
 }

//---------------------
// Erhht die Spannung
//---------------------
/*
void IncrementSoll (void)
 {
  if (SollPause > 1) { SollPuls++; SollPause--; }
 }      */

//---------------------------
// Setzt die Spannung runter
//---------------------------
/*
void DecrementSoll (void)
 {
  if (SollPuls > 1) { SollPuls--; SollPause++; }
 }
*/
//-----------------------------------
// Stellt direkt einen Sollwert ein
// Wertebereich 0 .. PulsMax-1 (Max)
//-----------------------------------
void SetSoll (char Wert)
 {
  SollPuls = Wert;
  SollPause = PulsMax-Wert;
 }

//-----------------------------------
// Stellt direkt einen prozentualen
// Sollwert ein
//-----------------------------------
void SetPower (char prozent)
 {
  if ((prozent / 4) <  PulsMax) SetSoll (prozent / 4); else SetSoll (PulsMax-1);
 }

//--------------------------------------
// Interruptroutine zum Ein/Ausschalten
// Wird aufgerufen im Nulldurchgang
//--------------------------------------
void SpannungISR (void)
 {

  if (Schaltzustand)
   {
     // Portpin auf Low, schaltet Motor ein
     SpannungEin ();
     Puls++;
     if (Puls >= SollPuls) { Puls = 0; Pause = 0; Schaltzustand = FALSE;}
   } else

   {
     // Portpin auf High, schaltet Motor aus
     SpannungAus ();
     Pause++;
     if (Pause >= SollPause) { Puls = 0; Pause = 0; Schaltzustand = TRUE;}
   }

 }


/**********************************************
 *
 * Ermittelt die momentane Drehzahl des Motors
 *
 **********************************************/
unsigned int getDrehzahl (void)
 {
  unsigned int counter = 0;

  sprintf (outpbuf, "H");
  SendString (outpbuf);

  // Warten auf eine negative Flanke
  while ((PORTC & 4) && (counter < 0x7FFF)) counter++;
  if (counter == 0xFFFF) return (0);

  sprintf (outpbuf, "L");
  SendString (outpbuf);

  // Warten, bis Signal wieder high
  while (!(PORTC & 4));

  sprintf (outpbuf, "H ");
  SendString (outpbuf);

  // Warten bis zur nchsten Negativen Flanke
  counter = 30000;
  while ((PORTC & 4) && (counter > 0)) counter--;

  return (counter);

 }

/*************************
 *
 * Zentrale ISR
 *
 *************************/
void interrupt ISR (void)
 {

  //---------------
  // Nulldurchgang
  //---------------
  if (INTF)
   {
     SpannungISR ();                                       // Motor steuern
     INTF = 0;                                             // Flag rcksetzen
     WDR ();
   }

  //-------------------
  // Capture-Ereigniss
  //-------------------
  if (CCP1IF)
   {
    // Drehzahl = CCPR1L + (int)CCPR1H*256;                 // Wert festhalten
    // Drehzahl = TMR1L + (unsigned int)TMR1H*256;             // Wert festhalten

		TMR1ON=0;		                                         // Timer 1 Stop
		TMR1L=0;		                                         // Timer 1L laden
		TMR1H=0;		                                         // Timer 1H laden
		TMR1ON=1;		                                         // Timer 1 Start
		TMR1IF = 0;                                          // Timer 1 starten
    CCP1IF = 0;                                          // Flag rcksetzen
   }

  //----------------------------------------------------------------
  // Timer-Interrupt. Wenn aufgetreten ist kein Capture vorgekommen
  //----------------------------------------------------------------
  if (TMR1IF == 1)
  	{
  		/*
  		TMR1ON=0;		                                         // Timer 1 Stop
  		TMR1L=0;		                                         // Timer 1L laden
  		TMR1H=0;		                                         // Timer 1H laden
  		TMR1ON=1;		                                         // Timer 1 Start
      Drehzahl = 0xFFFF;                                   // Wert auf Maximum setzen (Motor dreht offenbar nicht)
  		*/
  		TMR1IF = 0;                                          // Timer 1 starten
  		WDR ();
  	}

  //------------
  // Usart-ISR
  //------------
  if (RCIF == 1)
  	{
  	 RCIF = 0;
  	}

  return;
}

//-----------------------------
// Initialisierung des Systems
//-----------------------------
void init (void)
 {

   INTCON = 0xc0;		      // Global Interrupts enabled

   // Portpins klarmchen
   TRISA = 0b11111111;    // 1 = Eingang, 0 = Ausgang
   TRISB = 0b00111111;
   TRISC = 0b10111111;

   // Einstellen der Seriellen Schnittstelle
   SPBRG=0x07;		        // 19200 Baud (bei BRGH=0), 10 Mhz
   TXSTA=0x20;		        // 8-Bit, Transmit enabled, syncr. Mode, low Speed
   RCSTA=0x90;		        // serial Port enabled, 8-Bit
   RCIE=1;			          // Interrupt serielle Schnittstelle enable
   BRGH = 1;
   SPBRG = 32;		        // 19200 Baud (bei BRGH=1), 10 Mhz

   // Timer klarmachen
   T1CON = 0b00110001;    // Prescaler 1:8, minimaldrehzahl ca. 4 /sec.
   TMR1IE=1;		          // Timer1 overflow Int enabled
   TMR1IF = 0;            // Flag rcksetzen

   // Capture klarmachen
   // CCP1CON = 0b00000100;  // Capture auf jeder fallenden Flanke
   // CCP1IE = 1;            // Interrupt freischalten

   // Externen Interrupt klarmachen
   INTEDG = 1;            // Interrupt auf steigender Flanke
   PEIE = 1;              // Periphere Interrupts freischalten
   INTE = 1;              // Externen Interrupt einschalten

   // Startmeldung absetzen
   sprintf (outpbuf, "\n\rEsse 1.0\0");
   SendString (outpbuf);
   sprintf (outpbuf, "\n\r(C) Fingers Welt 2005\n\r\0");
   SendString (outpbuf);

 }

//--------------------------------------------------
// Berichtet ber den aktuellen Zustand des Systems
//--------------------------------------------------
void Statusmeldung (void)
 {
  #define Proportionalfaktor  1.7

  int Umdrehungen;
  int Regelgroesse;

  // Aktuelle Gren ermitteln
  Potiwert = GetPoti ();                         // Drehwinkel des Potis erfassen
  Regelgroesse = (Potiwert * 6) / 10;
  Potiwert = Regelgroesse + 32;
  Potiwert *= 13;
  Potiwert /= 10;
  Drehzahl = getDrehzahl ();                     // Motordrehzahl messen
  Umdrehungen = (Drehzahl / 300);

  Regelgroesse = Potiwert - Umdrehungen;
  if (Regelgroesse < 0) Regelgroesse = 0;
  if (Regelgroesse > 100) Regelgroesse = 100;

  sprintf (outpbuf, "%d %u %d %d\n\r", Potiwert, Drehzahl, Umdrehungen, Regelgroesse);
  SendString (outpbuf);

  SetPower ((char)Regelgroesse);

 }

/*****************
 *
 * Hauptprogramm
 *
 *****************/
void main (void)
 {
  init ();                                   // Alles startklar machen

  sprintf (outpbuf, "Und los !");
  SendString (outpbuf);

  while (1)
   {

    Statusmeldung ();                        // Wie ist die Lage ?

   }

 }

